/*****************************************************************************

        AvsFilterMaxDif16.h
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (AvsFilterMaxDif16_HEADER_INCLUDED)
#define	AvsFilterMaxDif16_HEADER_INCLUDED

#if defined (_MSC_VER)
	#pragma once
	#pragma warning (4 : 4250)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#define	WIN32_LEAN_AND_MEAN
#define	NOMINMAX
#define	NOGDI
#include <windows.h>
#include	"avisynth.h"
#include	"MTSlicer.h"
#include	"PlaneProcCbInterface.h"
#include	"PlaneProcessor.h"
#include	"types.h"



class AvsFilterMaxDif16
:	public GenericVideoFilter
,	public PlaneProcCbInterface
{

/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

public:

	typedef	AvsFilterMaxDif16	ThisType;

	explicit			AvsFilterMaxDif16 (::IScriptEnvironment *env_ptr, ::PClip src1_clip_sptr, ::PClip src2_clip_sptr, ::PClip ref_clip_sptr, int y, int u, int v);
	virtual			~AvsFilterMaxDif16 () {}

	virtual ::PVideoFrame __stdcall
						GetFrame (int n, ::IScriptEnvironment *env_ptr);



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

protected:

	// PlaneProcCbInterface
	virtual void	do_process_plane (::PVideoFrame &dst_sptr, int n, ::IScriptEnvironment &env, int plane_index, int plane_id);



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	class TaskDataGlobal
	{
	public:
		ThisType *		_this_ptr;
		int				_w;
		uint8_t *		_dst_msb_ptr;
		uint8_t *		_dst_lsb_ptr;
		const uint8_t*	_sr1_msb_ptr;
		const uint8_t*	_sr1_lsb_ptr;
		const uint8_t*	_sr2_msb_ptr;
		const uint8_t*	_sr2_lsb_ptr;
		const uint8_t*	_ref_msb_ptr;
		const uint8_t*	_ref_lsb_ptr;
		int				_stride_dst;
		int				_stride_sr1;
		int				_stride_sr2;
		int				_stride_ref;
	};

	typedef	MTSlicer <ThisType, TaskDataGlobal>	Slicer;

	void				process_subplane (Slicer::TaskData &td);

	::PClip			_src1_clip_sptr;
	::PClip			_src2_clip_sptr;
	::PClip			_ref_clip_sptr;
	PlaneProcessor	_plane_proc;



/*\\\ FORBIDDEN MEMBER FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

						AvsFilterMaxDif16 ();
						AvsFilterMaxDif16 (const AvsFilterMaxDif16 &other);
	AvsFilterMaxDif16 &
						operator = (const AvsFilterMaxDif16 &other);
	bool				operator == (const AvsFilterMaxDif16 &other) const;
	bool				operator != (const AvsFilterMaxDif16 &other) const;

};	// class AvsFilterMaxDif16



//#include	"AvsFilterMaxDif16.hpp"



#endif	// AvsFilterMaxDif16_HEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
