/*****************************************************************************

        AvsFilterResize16.cpp
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"conc/CritSec.h"
#include	"AvsFilterResize16.h"
#include	"FilterResize.h"
#include	"fnc.h"

#include	<algorithm>

#include	<cassert>
#include	<cctype>
#include	<cstring>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



AvsFilterResize16::AvsFilterResize16 (::IScriptEnvironment *env_ptr, ::PClip src_clip_sptr, int dst_width, int dst_height, double win_x, double win_y, double win_w, double win_h, std::string kernel_fnc, double kernel_scale_h, double kernel_scale_v, int taps, bool a1_flag, double a1, bool a2_flag, double a2, bool a3_flag, double a3, int kovrspl, bool norm_flag, bool preserve_center_flag, std::string cplace, int y, int u, int v, std::string kernel_fnc_h, std::string kernel_fnc_v, double kernel_total_h, double kernel_total_v, bool invks_flag, OptBool invks_h, OptBool invks_v, int invks_taps)
:	GenericVideoFilter (src_clip_sptr)
,	_src_clip_sptr (src_clip_sptr)
,	_src_width (vi.width)
,	_src_height (vi.height / 2)
,	_dst_width (dst_width)
,	_dst_height (dst_height)
,	_invks_taps (invks_taps)
,	_win_x (win_x)
,	_win_y (win_y)
,	_win_w (win_w)
,	_win_h (win_h)
,	_kernel_fnc_h ()
,	_kernel_fnc_v ()
,	_kernel_scale_h (kernel_scale_h)
,	_kernel_scale_v (kernel_scale_v)
,	_norm_val_h (kernel_total_h)
,	_norm_val_v (kernel_total_v)
,	_norm_flag (norm_flag)
,	_invks_h_flag (cumulate_flag (invks_flag, invks_h))
,	_invks_v_flag (cumulate_flag (invks_flag, invks_v))
,	_preserve_center_flag (preserve_center_flag)
,	_chroma_placement (conv_str_to_chroma_placement (*env_ptr, cplace))
,	_plane_proc (vi, *this)
,	_sse2_flag (false)
,	_filter_mutex ()
,	_filter_aptr_map ()
/*,	_kernel_arr ()*/
{
	// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
	// Parameter check

	if (! vi.IsPlanar ())
	{
		env_ptr->ThrowError ("Dither_resize16: input must be planar.");
	}
	if (! PlaneProcessor::check_stack16_height (vi))
	{
		env_ptr->ThrowError (
			"Dither_resize16: stacked MSB/LSB data must have\n"
			"a height multiple of twice the maximum subsampling."
		);
	}
	if (dst_width <= 0)
	{
		env_ptr->ThrowError ("Dither_resize16: target_width must be > 0.");
	}
	if (dst_height <= 0)
	{
		env_ptr->ThrowError ("Dither_resize16: target_height must be > 0.");
	}
	if (kernel_scale_h == 0)
	{
		env_ptr->ThrowError ("Dither_resize16: fh must be not null.");
	}
	if (kernel_scale_v == 0)
	{
		env_ptr->ThrowError ("Dither_resize16: fv must be not null.");
	}
	if (kernel_total_h < 0)
	{
		env_ptr->ThrowError ("Dither_resize16: totalh must be positive or null.");
	}
	if (kernel_total_v < 0)
	{
		env_ptr->ThrowError ("Dither_resize16: totalv must be positive or null.");
	}
	if (taps < 1 || taps > 128)
	{
		env_ptr->ThrowError ("Dither_resize16: taps must be in the 1-128 range.");
	}
	if (_invks_taps < 1 || _invks_taps > 128)
	{
		env_ptr->ThrowError ("Dither_resize16: invkstaps must be in the 1-128 range.");
	}

	if (_win_w == 0)
	{
		_win_w = _src_width;
	}
	else if (_win_w < 0)
	{
		_win_w = _src_width + _win_w - _win_x;
		if (_win_w <= 0)
		{
			env_ptr->ThrowError ("Dither_resize16: source width must be > 0.");
		}
	}

	if (_win_h == 0)
	{
		_win_h = _src_height;
	}
	if (_win_h <= 0)
	{
		_win_h = _src_height + _win_h - _win_y;
		if (_win_h <= 0)
		{
			env_ptr->ThrowError ("Dither_resize16: source height must be > 0.");
		}
	}

	_plane_proc.set_proc_mode (y, u, v);
	_plane_proc.set_clip_info (0, PlaneProcessor::ClipType_STACKED_16);
	_plane_proc.set_clip_info (1, PlaneProcessor::ClipType_STACKED_16);

	try
	{
		std::vector <double>	coef_arr;

		coef_arr.clear ();
		_kernel_fnc_h = kernel_fnc;
		if (! kernel_fnc_h.empty ())
		{
			_kernel_fnc_h = kernel_fnc_h;
		}
		_kernel_arr [FilterResize::Dir_H].create_kernel (
			_kernel_fnc_h, coef_arr, taps,
			a1_flag, a1,
			a2_flag, a2,
			a3_flag, a3,
			kovrspl,
			_invks_h_flag,
			_invks_taps
		);

		coef_arr.clear ();
		_kernel_fnc_v = kernel_fnc;
		if (! kernel_fnc_v.empty ())
		{
			_kernel_fnc_v = kernel_fnc_v;
		}
		_kernel_arr [FilterResize::Dir_V].create_kernel (
			_kernel_fnc_v, coef_arr, taps,
			a1_flag, a1,
			a2_flag, a2,
			a3_flag, a3,
			kovrspl,
			_invks_v_flag,
			_invks_taps
		);
	}
	catch (std::runtime_error &e)
	{
		static char		err_msg [4095+1];
		sprintf (err_msg, "Dither_resize16: %s", e.what ());
		env_ptr->ThrowError (err_msg);
	}
	catch (...)
	{
		env_ptr->ThrowError ("Dither_resize16: exception.");
	}

	vi.width  = _dst_width;
	vi.height = _dst_height * 2;

	if (! PlaneProcessor::check_stack16_width (vi))
	{
		env_ptr->ThrowError (
			"Dither_resize16: \"width\" must be\n"
			"multiple of the maximum subsampling."
		);
	}
	if (! PlaneProcessor::check_stack16_height (vi))
	{
		env_ptr->ThrowError (
			"Dither_resize16: \"height\" must be\n"
			"multiple of the maximum subsampling."
		);
	}

	init_fnc (*env_ptr);
}



::PVideoFrame __stdcall	AvsFilterResize16::GetFrame (int n, ::IScriptEnvironment *env_ptr)
{
	::PVideoFrame	dst_sptr = env_ptr->NewVideoFrame (vi);

	_plane_proc.process_frame (dst_sptr, n, *env_ptr, &_src_clip_sptr, 0, 0);

	return (dst_sptr);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	AvsFilterResize16::do_process_plane (::PVideoFrame &dst_sptr, int n, ::IScriptEnvironment &env, int plane_index, int plane_id)
{
	::PVideoFrame	src_sptr = _src_clip_sptr->GetFrame (n, &env);

	const uint8_t*	data_src_ptr = src_sptr->GetReadPtr (plane_id);
	uint8_t *		data_dst_ptr = dst_sptr->GetWritePtr (plane_id);
	const int		dst_w        = _plane_proc.get_width (dst_sptr, plane_id);
	const int		dst_h        = _plane_proc.get_height16 (dst_sptr, plane_id);
	const int		src_w        = _plane_proc.get_width (src_sptr, plane_id);
	const int		src_h        = _plane_proc.get_height16 (src_sptr, plane_id);
	const int		stride_dst   = dst_sptr->GetPitch (plane_id);
	const int		stride_src   = src_sptr->GetPitch (plane_id);

	const int		subspl_h = vi.width  / dst_w;
	const int		subspl_v = vi.height / (dst_h * 2);

	const int		offset_lsb_dst = dst_h * stride_dst;
	const int		offset_lsb_src = src_h * stride_src;

	// http://www.mir.com/DMG/chroma.html

	// center_pos is the position of the sampling point relative to the frame
	// top/left border, in the plane coordinates. For reference, the border
	// of the frame is at 0.5 units of luma from the first luma sampling point.
	// I. e., the luma sampling point is at the pixel's center.

	// Generic case for luma, non-subsampled chroma and MPEG-1 chroma.
	double			center_pos_h = 0.5;
	double			center_pos_v = 0.5;

	if (vi.IsFieldBased () && vi.IsParityKnown ())
	{
		center_pos_v *= 0.5;
		if (   (vi.IsTFF () && (n & 1) == 1)
		    || (vi.IsBFF () && (n & 1) == 0))
		{
			center_pos_v += 0.5;
		}
	}

	// Subsampled chroma
	if (plane_id != PLANAR_Y)
	{
		if (subspl_h > 1)
		{
			if (   _chroma_placement == ChromaPlacement_MPEG2
			    || _chroma_placement == ChromaPlacement_DV)
			{
				center_pos_h = 0.5 / subspl_h;
			}
		}

		if (subspl_v == 2)
		{
			if (_chroma_placement == ChromaPlacement_MPEG2)
			{
				if (vi.IsFieldBased () && vi.IsParityKnown ())
				{
					center_pos_v = 0.25;
					if (   (vi.IsTFF () && (n & 1) == 1)
					    || (vi.IsBFF () && (n & 1) == 0))
					{
						center_pos_v = 0.75;
					}
				}
				else
				{
					center_pos_v = 0.5;
				}
			}
			else if (_chroma_placement == ChromaPlacement_DV)
			{
				center_pos_v = 0.25;

				if (vi.IsFieldBased () && vi.IsParityKnown ())
				{
					center_pos_v *= 0.5;
					if (   (vi.IsTFF () && (n & 1) == 1)
					    || (vi.IsBFF () && (n & 1) == 0))
					{
						center_pos_v += 0.25;
					}
				}

				if (plane_id == PLANAR_V)
				{
					center_pos_v += 0.5;
				}
			}
		}
	}

	try
	{
		FilterResize *	filter_ptr = create_or_access_plane_filter (
			plane_id,
			src_w, src_h,
			dst_w, dst_h,
			_win_x / subspl_h, _win_y / subspl_v,
			_win_w / subspl_h, _win_h / subspl_v,
			(_preserve_center_flag) ? center_pos_h : 0,
			(_preserve_center_flag) ? center_pos_v : 0
		);

		filter_ptr->process_plane (
			data_dst_ptr, data_dst_ptr + offset_lsb_dst,
			data_src_ptr, data_src_ptr + offset_lsb_src,
			stride_dst,
			stride_src
		);
	}

	catch (std::exception &e)
	{
		env.ThrowError (e.what ());
	}
}



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	AvsFilterResize16::init_fnc (::IScriptEnvironment &env)
{
	const long		cpu_flags = env.GetCPUFlags ();
	if (   (cpu_flags & CPUF_SSE2) != 0
	    && (cpu_flags & CPUF_SSE ) != 0)
	{
		_sse2_flag = true;
	}
}



FilterResize *	AvsFilterResize16::create_or_access_plane_filter (int plane_id, int src_width, int src_height, int dst_width, int dst_height, double win_x, double win_y, double win_w, double win_h, double center_pos_h, double center_pos_v)
{
	assert (src_width > 0);
	assert (src_height > 0);
	assert (dst_width > 0);
	assert (dst_height > 0);
	assert (win_w > 0);
	assert (win_h > 0);

	FilterResize * filter_ptr = 0;
	const int64_t  key1 = (int64_t (src_height) << 32) + src_width;
	const int64_t  key2 =   (int64_t (round_int (center_pos_v * 64)) << 48)
	                      + (int64_t (round_int (center_pos_h * 64)) << 16);
	const int64_t  key  = key1 ^ key2;

	conc::CritSec  autolock (_filter_mutex);

	SharedPtr <FilterResize> &	filter_aptr = _filter_aptr_map [key];
	if (filter_aptr.get () == 0)
	{
		filter_aptr = SharedPtr <FilterResize> (new FilterResize (
			src_width, src_height, dst_width, dst_height,
			win_x, win_y, win_w, win_h,
			*(_kernel_arr [FilterResize::Dir_H]._k_aptr),
			*(_kernel_arr [FilterResize::Dir_V]._k_aptr),
			_kernel_scale_h, _kernel_scale_v,
			_norm_flag, _norm_val_h, _norm_val_v,
			center_pos_h, center_pos_v,
			_sse2_flag
		));
	}

	return (filter_aptr.get ());
}



AvsFilterResize16::ChromaPlacement	AvsFilterResize16::conv_str_to_chroma_placement (::IScriptEnvironment &env, std::string cplace)
{
	assert (&env != 0);
	assert (&cplace != 0);

	ChromaPlacement	cp_val = ChromaPlacement_MPEG1;

	conv_to_lower_case (cplace);
	if (strcmp (cplace.c_str (), "mpeg1") == 0)
	{
		cp_val = ChromaPlacement_MPEG1;
	}
	else if (strcmp (cplace.c_str (), "mpeg2") == 0)
	{
		cp_val = ChromaPlacement_MPEG2;
	}
	else if (strcmp (cplace.c_str (), "dv") == 0)
	{
		cp_val = ChromaPlacement_DV;
	}
	else
	{
		env.ThrowError ("Dither_resize16: unexpected cplace string.");
	}

	return (cp_val);
}



bool	AvsFilterResize16::cumulate_flag (bool flag, OptBool flag2)
{
	if (flag2 == OptBool_TRUE)
	{
		flag = true;
	}
	else if (flag2 == OptBool_FALSE)
	{
		flag = false;
	}
	else
	{
		assert (flag2 == OptBool_UNDEF);
	}

	return (flag);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
