/*****************************************************************************

        KernelData.h
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (KernelData_HEADER_INCLUDED)
#define	KernelData_HEADER_INCLUDED

#if defined (_MSC_VER)
	#pragma once
	#pragma warning (4 : 4250)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"ContFirInterface.h"
#include	"DiscreteFirInterface.h"

#include	<memory>
#include	<string>
#include	<vector>



class KernelData
{

/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

public:

	               KernelData ();
	virtual        ~KernelData () {}

	void           create_kernel (std::string kernel_fnc, std::vector <double> &coef_arr, int taps, bool a1_flag, double a1, bool a2_flag, double a2, bool a3_flag, double a3, int kovrspl, bool inv_flag, int inv_taps);
	std::auto_ptr <ContFirInterface>
	               _k_aptr;
	std::auto_ptr <DiscreteFirInterface>
	               _discrete_aptr;



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

protected:



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	void           create_kernel_base (std::string kernel_fnc, std::vector <double> &coef_arr, int taps, bool a1_flag, double a1, bool a2_flag, double a2, bool a3_flag, double a3, int kovrspl);
	void           invert_kernel (int taps);

	static void    conv_to_float_arr (std::vector <double> &coef_arr, const std::string &str);

	template <class W>
	static void    apply_window (std::vector <double> &x, double &norm_sum, int taps, int h_len, int h_len_i, double inv_ovr_s);



/*\\\ FORBIDDEN MEMBER FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	               KernelData (const KernelData &other);
	KernelData &   operator = (const KernelData &other);
	bool           operator == (const KernelData &other) const;
	bool           operator != (const KernelData &other) const;

};	// class KernelData



//#include	"KernelData.hpp"



#endif	// KernelData_HEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
