/*****************************************************************************

        ToolsSse2.hpp
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (ToolsSse2_CODEHEADER_INCLUDED)
#define	ToolsSse2_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	<cassert>
#include	<cstddef>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



__m128i	ToolsSse2::loada_8_16ml (const void *msb_ptr, const void *lsb_ptr)
{
	assert (msb_ptr != 0);
	assert (lsb_ptr != 0);

	const __m128i	val_msb = _mm_loadl_epi64 (
		reinterpret_cast <const __m128i *> (msb_ptr)
	);
	const __m128i	val_lsb = _mm_loadl_epi64 (
		reinterpret_cast <const __m128i *> (lsb_ptr)
	);
	const __m128i	val = _mm_unpacklo_epi8 (val_lsb, val_msb);

	return (val);
}



// mask_lsb = 0x00FF00FF00FF00FF00FF00FF00FF00FF
void	ToolsSse2::storea_8_16ml (void *msb_ptr, void *lsb_ptr, __m128i val, __m128i mask_lsb)
{
	assert (msb_ptr != 0);
	assert (lsb_ptr != 0);
	assert (lsb_ptr != msb_ptr);

	const __m128i	lsb = _mm_and_si128 (mask_lsb, val);
	__m128i			msb = _mm_andnot_si128 (mask_lsb, val);
	msb = _mm_srli_si128 (msb, 1);

	__m128i			tmp = _mm_packus_epi16 (lsb, msb);
	_mm_storel_epi64 (reinterpret_cast <__m128i *> (lsb_ptr), tmp);

	tmp = _mm_unpackhi_epi64 (tmp, tmp);
	_mm_storel_epi64 (reinterpret_cast <__m128i *> (msb_ptr), tmp);
}



// mask_store = 0x0000...000080...808080
void	ToolsSse2::storea_8_16ml_partial (void *msb_ptr, void *lsb_ptr, __m128i val, __m128i mask_lsb, __m128i mask_store)
{
	assert (msb_ptr != 0);
	assert (lsb_ptr != 0);
	assert (lsb_ptr != msb_ptr);

	const __m128i	lsb = _mm_and_si128 (mask_lsb, val);
	__m128i			msb = _mm_andnot_si128 (mask_lsb, val);
	msb = _mm_srli_si128 (msb, 1);

	__m128i			tmp = _mm_packus_epi16 (lsb, msb);
	_mm_maskmoveu_si128 (tmp, mask_store, reinterpret_cast <char *> (lsb_ptr));

	tmp = _mm_unpackhi_epi64 (tmp, tmp);
	_mm_maskmoveu_si128 (tmp, mask_store, reinterpret_cast <char *> (msb_ptr));
}



__m128i	ToolsSse2::create_store_mask (int nbr_pix)
{
	assert (nbr_pix >= 0);
	assert (nbr_pix < 8);

	const int      v0 = (nbr_pix >= 4) ? -1 : 0x00FFFFFF >> ((3 - nbr_pix) * 8);
	const int      v1 = (nbr_pix <  4) ?  0 : 0x00FFFFFF >> ((7 - nbr_pix) * 8);
	const __m128i  mask_store = _mm_set_epi32 (0, 0, v1, v0);

	return (mask_store);
}



// Code of this function shamelessly borrowed from tp7
// https://github.com/tp7/masktools/blob/16bit/masktools/common/simd.h
__forceinline __m128i	ToolsSse2::mullo_epi32 (const __m128i &a, const __m128i &b)
{
	const __m128i  a13    = _mm_shuffle_epi32 (a, 0xF5);         // (-,a3,-,a1)
	const __m128i  b13    = _mm_shuffle_epi32 (b, 0xF5);         // (-,b3,-,b1)
	const __m128i  prod02 = _mm_mul_epu32 (a, b);                // (-,a2*b2,-,a0*b0)
	const __m128i  prod13 = _mm_mul_epu32 (a13, b13);            // (-,a3*b3,-,a1*b1)
	const __m128i  prod01 = _mm_unpacklo_epi32 (prod02, prod13); // (-,-,a1*b1,a0*b0)
	const __m128i  prod23 = _mm_unpackhi_epi32 (prod02, prod13); // (-,-,a3*b3,a2*b2)
	const __m128i  res    = _mm_unpacklo_epi64 (prod01 ,prod23); // (ab3,ab2,ab1,ab0)

	return (res);
}



// Same here
__forceinline __m128i	ToolsSse2::pack_epi16 (__m128i a, __m128i b)
{
#if 0	// To do: test which one is faster

	a = _mm_shufflelo_epi16 (a, _MM_SHUFFLE (3, 3, 2, 0));
	a = _mm_shufflehi_epi16 (a, _MM_SHUFFLE (3, 3, 2, 0));
	a = _mm_shuffle_epi32 (a, _MM_SHUFFLE (3, 3, 2, 0));

	b = _mm_shufflelo_epi16 (b, _MM_SHUFFLE (3, 3, 2, 0));
	b = _mm_shufflehi_epi16 (b, _MM_SHUFFLE (3, 3, 2, 0));
	b = _mm_shuffle_epi32 (b, _MM_SHUFFLE (2, 0, 3, 3));

	const __m128i  ma = _mm_set_epi32 (-1, -1, 0, 0);
	const __m128i  pa = _mm_and_si128 (ma, a);
	const __m128i  pb = _mm_andnot_si128 (ma, b);
	const __m128i  p  = _mm_or_si128 (pa, pb);

#else

	const __m128i  mask_lo  = _mm_load_si128 (
		reinterpret_cast <const __m128i *> (ToolsSse2::_c32_0000ffff)
	);
	const __m128i  offset_s = _mm_load_si128 (
		reinterpret_cast <const __m128i *> (ToolsSse2::_c32_00008000)
	);
	const __m128i  mask_s   = _mm_load_si128 (
		reinterpret_cast <const __m128i *> (ToolsSse2::_c16_8000)
	);

	a = _mm_and_si128 (mask_lo, a);
	b = _mm_and_si128 (mask_lo, b);
	a = _mm_sub_epi32 (a, offset_s);
	b = _mm_sub_epi32 (b, offset_s);
	__m128i        p = _mm_packs_epi32 (a, b);
	p = _mm_xor_si128 (p, mask_s);

#endif

	return (p);
}



__forceinline __m128i	ToolsSse2::select (const __m128i &cond, const __m128i &v_t, const __m128i &v_f)
{
	const __m128i  cond_1   = _mm_and_si128 (cond, v_t);
	const __m128i  cond_0   = _mm_andnot_si128 (cond, v_f);
	const __m128i  res      = _mm_or_si128 (cond_0, cond_1);

	return (res);
}



template <class T>
bool	ToolsSse2::check_ptr_align (T *ptr)
{
	return (ptr != 0 && (reinterpret_cast <ptrdiff_t> (ptr) & 15) == 0);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



#endif	// ToolsSse2_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
